import json
from dataclasses import dataclass, field

from tarzan.features.text import Text


@dataclass
class Json(Text):
    # Automatically constructed
    _type: str = field(default="Json", init=False, repr=False)

    def encode_example(self, example):
        return super().encode_example(json.dumps(example))

    def decode_example(self, example):
        return json.loads(super().decode_example(example))



from dataclasses import dataclass, field
from typing import Any, ClassVar


@dataclass
class Sequence:
    """Construct a list of feature from a single type or a dict of types.
    Mostly here for compatiblity with tfds.

    Args:
        feature:
            A list of features of a single type or a dictionary of types.
        length (`int`):
            Length of the sequence.

    Example:

    ```py
    >>> from torchtts.data.v2.features import Features, Sequence, Text, Scalar
    >>> features = Features({'post': Sequence(feature={'text': Text(), 'upvotes': Scalar(dtype='int32')})})
    >>> features
    {'post': Sequence(feature={'text': Text(), 'upvotes': Scalar(shape=(), dtype='int32')}, length=-1)}
    ```
    """

    feature: Any
    length: int = -1
    # Automatically constructed
    dtype: ClassVar[str] = "list"
    _type: str = field(default="Sequence", init=False, repr=False)



import copy
import dataclasses
import json
import logging
from dataclasses import dataclass
from typing import Dict, List, Optional

from tarzan.features.features import Features
from tarzan.utils import asdict, update_dict

logger = logging.getLogger(__name__)


@dataclass
class DatasetInfo:
    """Information about a dataset."""

    description: str = dataclasses.field(default_factory=str)
    file_list: List[str] = dataclasses.field(default_factory=list)
    features: Optional[Features] = None
    size_in_bytes: Optional[int] = None
    metadata: Optional[Dict] = None

    def __post_init__(self):
        # Convert back to the correct classes when we reload from dict
        if self.features is not None and not isinstance(self.features, Features):
            self.features = Features.from_dict(self.features)

    def write_to_json(self, dataset_info_file, pretty_print=False):
        """Write `DatasetInfo` and license (if present) as JSON files to `dataset_info_dir`.

        Args:
            dataset_info_file (`str`):
                Destination json file.
            pretty_print (`bool`, defaults to `False`):
                If `True`, the JSON will be pretty-printed with the indent level of 4.
        """
        if self.file_list is None:
            logger.warning(
                "No file list provided, the dataset info will be incomplete."
            )
        with open(dataset_info_file, "wb") as f:
            self._dump_info(f, pretty_print=pretty_print)

    def _dump_info(self, file, pretty_print=False):
        """Dump info in `file` file-like object open in bytes mode (to support remote files)"""
        file.write(
            json.dumps(asdict(self), indent=4 if pretty_print else None).encode("utf-8")
        )

    @classmethod
    def from_json(cls, dataset_info_file: str) -> "DatasetInfo":
        """Create [`DatasetInfo`] from the JSON file `dataset_info_file`.

        This function updates all the dynamically generated fields (num_examples,
        hash, time of creation,...) of the [`DatasetInfo`].

        This will overwrite all previous metadata.

        Args:
            dataset_info_file (`str`):
                The Json file of dataset info.
        """
        logger.info(f"Loading Dataset info from {dataset_info_file}")
        with open(dataset_info_file, "r", encoding="utf-8") as f:
            dataset_info_dict = json.load(f)
        return cls.from_dict(dataset_info_dict)

    @classmethod
    def from_dict(cls, dataset_info_dict: Dict) -> "DatasetInfo":
        field_names = {f.name for f in dataclasses.fields(cls)}
        return cls(**{k: v for k, v in dataset_info_dict.items() if k in field_names})

    def update(self, other_dataset_info: "DatasetInfo", ignore_none=True):
        self_dict = self.__dict__
        self.__dict__ = update_dict(
            self_dict,
            {
                k: copy.deepcopy(v)
                for k, v in other_dataset_info.__dict__.items()
                if (v is not None or not ignore_none)
            },
        )

    def copy(self) -> "DatasetInfo":
        return self.__class__(**{k: copy.deepcopy(v) for k, v in self.__dict__.items()})



import collections
import copy
import itertools
from dataclasses import fields, is_dataclass
from typing import Dict, Any


def zip_dict(*dicts):
    """Iterate over items of dictionaries grouped by their keys."""
    for key in set(itertools.chain(*dicts)):  # set merge all keys
        # Will raise KeyError if the dict don't have the same keys
        yield key, tuple(d[key] for d in dicts)


def update_dict(d, u):
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = update_dict(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def asdict(obj):
    """Convert an object to its dictionary representation recursively.

    <Added version="2.4.0"/>
    """

    # Implementation based on https://docs.python.org/3/library/dataclasses.html#dataclasses.asdict

    def _is_dataclass_instance(obj):
        # https://docs.python.org/3/library/dataclasses.html#dataclasses.is_dataclass
        return is_dataclass(obj) and not isinstance(obj, type)

    def _asdict_inner(obj):
        if _is_dataclass_instance(obj):
            result = {}
            for f in fields(obj):
                value = _asdict_inner(getattr(obj, f.name))
                if (
                    not f.init
                    or value != f.default
                    or f.metadata.get("include_in_asdict_even_if_is_default", False)
                ):
                    result[f.name] = value
            return result
        elif isinstance(obj, tuple) and hasattr(obj, "_fields"):
            # obj is a namedtuple
            return type(obj)(*[_asdict_inner(v) for v in obj])
        elif isinstance(obj, (list, tuple)):
            # Assume we can create an object of this type by passing in a
            # generator (which is not true for namedtuples, handled
            # above).
            return type(obj)(_asdict_inner(v) for v in obj)
        elif isinstance(obj, dict):
            return {_asdict_inner(k): _asdict_inner(v) for k, v in obj.items()}
        else:
            return copy.deepcopy(obj)

    if not isinstance(obj, dict) and not _is_dataclass_instance(obj):
        raise TypeError(f"{obj} is not a dict or a dataclass")

    return _asdict_inner(obj)


class StreamWrapper:
    """
    StreamWrapper is introduced to wrap file handler generated by DataPipe operation like `FileOpener`.

    StreamWrapper would guarantee the wrapped file handler is closed when it's out of scope.
    """

    session_streams: Dict[Any, int] = {}
    debug_unclosed_streams: bool = False

    def __init__(self, file_obj, parent_stream=None, name=None):
        self.file_obj = file_obj
        self.child_counter = 0
        self.parent_stream = parent_stream
        self.close_on_last_child = False
        self.name = name
        self.closed = False
        if parent_stream is not None:
            if not isinstance(parent_stream, StreamWrapper):
                raise RuntimeError(f'Parent stream should be StreamWrapper, {type(parent_stream)} was given')
            parent_stream.child_counter += 1
            self.parent_stream = parent_stream
        if StreamWrapper.debug_unclosed_streams:
            StreamWrapper.session_streams[self] = 1

    @classmethod
    def close_streams(cls, v, depth=0):
        """Traverse structure and attempts to close all found StreamWrappers on best effort basis."""
        if depth > 10:
            return
        if isinstance(v, StreamWrapper):
            v.close()
        else:
            # Traverse only simple structures
            if isinstance(v, dict):
                for vv in v.values():
                    cls.close_streams(vv, depth=depth + 1)
            elif isinstance(v, (list, tuple)):
                for vv in v:
                    cls.close_streams(vv, depth=depth + 1)

    def __getattr__(self, name):
        file_obj = self.__dict__['file_obj']
        return getattr(file_obj, name)

    def close(self, *args, **kwargs):
        if self.closed:
            return
        if StreamWrapper.debug_unclosed_streams:
            del StreamWrapper.session_streams[self]
        if hasattr(self, "parent_stream") and self.parent_stream is not None:
            self.parent_stream.child_counter -= 1
            if not self.parent_stream.child_counter and self.parent_stream.close_on_last_child:
                self.parent_stream.close()
        try:
            self.file_obj.close(*args, **kwargs)
        except AttributeError:
            pass
        self.closed = True

    def autoclose(self):
        """Automatically close stream when all child streams are closed or if there are none."""
        self.close_on_last_child = True
        if self.child_counter == 0:
            self.close()

    def __dir__(self):
        attrs = list(self.__dict__.keys()) + list(StreamWrapper.__dict__.keys())
        attrs += dir(self.file_obj)
        return list(set(attrs))

    def __del__(self):
        if not self.closed:
            self.close()

    def __iter__(self):
        yield from self.file_obj

    def __next__(self):
        return next(self.file_obj)

    def __repr__(self):
        if self.name is None:
            return f"StreamWrapper<{self.file_obj!r}>"
        else:
            return f"StreamWrapper<{self.name},{self.file_obj!r}>"

    def __getstate__(self):
        return self.file_obj

    def __setstate__(self, obj):
        self.file_obj = obj



from dataclasses import dataclass, field

from tarzan.features.tensor import Tensor


@dataclass
class Scalar(Tensor):
    # Automatically constructed
    _type: str = field(default="Scalar", init=False, repr=False)

    def __init__(self, dtype, **kwargs):
        super().__init__(shape=(), dtype=dtype)



from dataclasses import dataclass, field

from tarzan.utils import StreamWrapper


@dataclass
class Text:
    # Automatically constructed
    _type: str = field(default="Text", init=False, repr=False)

    def encode_example(self, example):
        return example.encode("utf-8")

    def decode_example(self, example):
        if isinstance(example, StreamWrapper):
            fobj = example
            example = example.read()
            fobj.close()
            if not example:
                return None
        return example.decode("utf-8")



from abc import ABC, abstractmethod
from typing import Any


class Writer(ABC):
    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.close()

    @abstractmethod
    def write(self, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def close(self):
        raise NotImplementedError



import io
import logging
import os
from dataclasses import dataclass, field
from typing import Optional

import librosa
import numpy as np

from tarzan.features.tensor import Tensor

logger = logging.getLogger(__name__)


class AudioDecoder:
    """Read audio during decoding."""

    def __init__(self, fobj, dtype, shape, sample_rate):
        self._fobj = fobj
        self._dtype = dtype
        self._shape = shape
        self._sample_rate = sample_rate
        self._channels = shape[1] if len(shape) > 1 else 1

    def read_all(self):
        self._fobj.seek(0)
        try:
            return librosa.load(
                self._fobj, sr=self._sample_rate, mono=self._channels == 1, dtype=self._dtype
            )
        except Exception as e:
            logger.error(f"Error reading audio: {e}")
            return None, None

    def read_range(self, start: float, end: float):
        if start < 0:
            raise ValueError(f"start must be >= 0, got {start}")
        if start > end:
            raise ValueError(f"end must be >= start, got {end} < {start}")
        duration = end - start
        self._fobj.seek(0)
        try:
            return librosa.load(
                self._fobj,
                sr=self._sample_rate,
                mono=self._channels == 1,
                dtype=self._dtype,
                offset=start,
                duration=duration,
            )
        except Exception as e:
            logger.error(f"Error reading audio: {e}")
            return None, None


@dataclass
class Audio(Tensor):
    """`FeatureConnector` for audio."""

    sample_rate: Optional[int]
    # Automatically constructed
    _type: str = field(default="Audio", init=False, repr=False)

    def __init__(
        self, shape=(None,), dtype="float32", sample_rate=None, lazy_decode=True
    ):
        super().__init__(shape=shape, dtype=dtype)
        self.sample_rate = sample_rate

    def encode_example(self, audio_or_path_or_fobj):
        if isinstance(audio_or_path_or_fobj, (str, os.PathLike)):
            filename = os.fspath(audio_or_path_or_fobj)
            with open(filename, "rb") as audio_f:
                audio = audio_f.read()
        elif isinstance(audio_or_path_or_fobj, np.ndarray):
            raise ValueError("Audio must be a path or file-like object.")
        else:
            audio = audio_or_path_or_fobj.read()
        return audio

    def decode_example(self, audio_data):
        if isinstance(audio_data, bytes):
            audio_fobj = io.BytesIO(audio_data)
        else:
            audio_fobj = audio_data
        return AudioDecoder(audio_fobj, self.dtype, self.shape, self.sample_rate)



import logging
import os
import tarfile
from typing import Dict, Iterator, List, Optional, Tuple

from more_itertools import peekable

from tarzan.features import Features
from tarzan.info import DatasetInfo
from tarzan.utils import StreamWrapper

logger = logging.getLogger(__name__)


def _get_tar_index(tarinfo: tarfile.TarInfo):
    return tarinfo.name.split("/")[0]


def _feature_stream(stream: StreamWrapper):
    peekable_tar = peekable(stream)

    feature_group = []

    while True:
        # Group successive features with same index prefix
        try:
            tarinfo = next(peekable_tar)
        except StopIteration:
            break
        index = _get_tar_index(tarinfo)
        feature_group.append(tarinfo)

        try:
            tarinfo = peekable_tar.peek()
        # Last index case
        except StopIteration:
            yield index, _compose_feature(stream, feature_group)
            break

        # Successive index case
        if _get_tar_index(tarinfo) != index:
            yield index, _compose_feature(stream, feature_group)
            feature_group.clear()


def _get_inner_fobj(tar_stream: StreamWrapper, tarinfo: tarfile.TarInfo):
    extracted_fobj = tar_stream.extractfile(tarinfo)
    if extracted_fobj is None:
        raise tarfile.ExtractError(
            f"failed to extract file {tarinfo.name} from {tar_stream.name}"
        )
    inner_pathname = os.path.normpath(os.path.join(tar_stream.name, tarinfo.name))
    return StreamWrapper(extracted_fobj, tar_stream, name=inner_pathname)


def _compose_feature(tar_stream: StreamWrapper, group: List[tarfile.TarInfo]):
    """Compose feature (maybe nested) with same index prefix."""
    # Single feature case
    if len(group) == 1:
        tarinfo = group[0]
        # Single feature must be a file
        assert tarinfo.isfile()
        return _get_inner_fobj(tar_stream, tarinfo)
    # Nested feature case
    else:
        nested = {}
        for tarinfo in group:
            if tarinfo.isdir():
                continue
            parts = tarinfo.name.split("/")
            current_dict = nested
            for part in parts[1:-1]:
                if part not in current_dict:
                    current_dict[part] = {}
                current_dict = current_dict[part]
            current_dict[parts[-1]] = _get_inner_fobj(tar_stream, tarinfo)
        return _transform_dict(nested)


def _transform_dict(input_dict):
    """Transform a nested dictionary to a list if all keys are integers."""
    result_dict = {}
    for key, value in input_dict.items():
        if isinstance(value, dict):
            value = _transform_dict(value)
        result_dict[key] = value
    if all(key.isdigit() for key in result_dict.keys()):
        result_dict = [
            value
            for key, value in sorted(result_dict.items(), key=lambda item: int(item[0]))
        ]

    return result_dict


class TarReader:
    def __init__(
        self,
        tar_files: List[str],
        features: Features,
        mode: str = "r:*",
    ) -> None:
        super().__init__()
        self.tar_files = tar_files
        self.features = features
        self.mode = mode

    @classmethod
    def from_dataset_info(cls, dataset_info_file: str) -> "TarReader":
        info = DatasetInfo.from_json(dataset_info_file)
        data_dir = os.path.dirname(dataset_info_file)
        tar_files = [os.path.join(data_dir, f) for f in info.file_list]
        return cls(tar_files=tar_files, features=info.features)

    def __iter__(self) -> Iterator[Tuple[str, str, Dict]]:
        for tar_file in self.tar_files:
            tar_stream = StreamWrapper(
                tarfile.open(tar_file, self.mode),
                name=tar_file,
            )
            example_stream = _feature_stream(tar_stream)

            try:
                while True:
                    try:
                        index, example = next(example_stream)
                    except StopIteration:
                        break
                    decoded_example = self.features.decode_example(example)
                    yield tar_file, index, decoded_example
            finally:
                tar_stream.autoclose()



import io
import logging
import os.path
import tarfile
from typing import Any, Dict

from tarzan.features import Features
from tarzan.writers.base_writer import Writer

logger = logging.getLogger(__name__)


def _add_dir_to_tar(tar, name):
    tar_info = tarfile.TarInfo(name=name)
    tar_info.type = tarfile.DIRTYPE
    tar_info.mode = 0o755
    tar.addfile(tar_info)


def _write_to_tar(tar, prefix, data):
    size = 0
    if isinstance(data, dict):
        _add_dir_to_tar(tar, prefix)
        for key, value in data.items():
            if key.isdigit():
                raise ValueError(
                    "Keys cannot be integers since they are reserved for indexing."
                )
            size += _write_to_tar(tar, f"{prefix}/{key}", value)
    elif isinstance(data, list):
        _add_dir_to_tar(tar, prefix)
        for i, value in enumerate(data):
            size += _write_to_tar(tar, f"{prefix}/{i}", value)
    else:
        # Assumed to be bytes or None
        tarinfo = tarfile.TarInfo(name=prefix)
        if data is None:
            size = 0
            tar.addfile(tarinfo)
        else:
            size = len(data)
            tarinfo.size = len(data)
            tar.addfile(tarinfo, fileobj=io.BytesIO(data))
    return size


class TarWriter(Writer):
    """TarWriter writes data to tar files with nested directory structure."""

    def __init__(
        self,
        path: str,
        features: Features,
    ):
        self.tar_stream = tarfile.open(path, "w|")
        self.features = features
        self.written_idx = set()

    def write(self, idx: str, objects: Dict[str, Any]):
        if objects.keys() != self.features.keys():
            raise ValueError(
                f"Keys {objects.keys()} do not match specified features {self.features}"
            )
        if idx in self.written_idx:
            raise ValueError(f"Index {idx} already written")
        written_obj = self.features.encode_example(objects)
        size = _write_to_tar(self.tar_stream, idx, written_obj)
        self.written_idx.add(idx)
        return size

    def close(self):
        self.tar_stream.close()



import logging
import os

from tarzan.info import DatasetInfo
from tarzan.writers.base_writer import Writer
from tarzan.writers.tar_writer import TarWriter

logger = logging.getLogger(__name__)


class ShardWriter(Writer):
    """Writer wrapper to split data into multiple shards."""
    DATASET_INFO_FILENAME = "dataset_info.json"

    def __init__(self,
                 path: str,
                 info: DatasetInfo,
                 pattern: str = "%05d",
                 max_count: int = 1000,
                 max_size: int = 3e9):
        self.path = path
        os.makedirs(path, exist_ok=True)

        self.pattern = f"{path}/{pattern}.tar"
        self.info = info
        self.max_count = max_count
        self.max_size = max_size

        self.writer_stream = None
        self.shard = 0
        self.count = 0
        self.size = 0
        self.total_count = 0
        self.fname = None
        self.next_stream()

    def next_stream(self):
        self.finish()
        self.fname = self.pattern % self.shard
        self.shard += 1
        self.writer_stream = TarWriter(self.fname, features=self.info.features)
        self.count = 0
        self.size = 0

    def write(self, obj):
        if self.writer_stream is None or self.count >= self.max_count or self.size > self.max_size:
            self.next_stream()
        size = self.writer_stream.write(str(self.count), obj)
        self.count += 1
        self.size += size
        self.total_count += 1

    def finish(self):
        if self.writer_stream is not None:
            self.writer_stream.close()
            assert self.fname is not None
            self.info.file_list.append(os.path.basename(self.fname))
            self.writer_stream = None

    def close(self):
        logger.info(f"{self.total_count} examples have been written to {self.shard} shards")
        self.finish()
        self.info.write_to_json(f"{self.path}/{self.DATASET_INFO_FILENAME}", pretty_print=True)



from typing import Union

from tarzan.features.audio import Audio
from tarzan.features.json import Json
from tarzan.features.scalar import Scalar
from tarzan.features.sequence import Sequence
from tarzan.features.tensor import Tensor
from tarzan.features.text import Text

FeatureType = Union[
    dict,
    list,
    tuple,
    Audio,
    Json,
    Scalar,
    Sequence,
    Text,
    Tensor,
]

from tarzan.features.features import Features



from dataclasses import dataclass, field

import numpy as np
from tarzan.utils import StreamWrapper


@dataclass
class Tensor:
    """Generic data of arbitrary shape and type."""

    shape: tuple
    dtype: str
    # Automatically constructed
    _type: str = field(default="Tensor", init=False, repr=False)

    def __post_init__(self):
        self.shape = tuple(self.shape)
        if not isinstance(self.dtype, str):
            raise TypeError("dtype must be a string")
        if not is_valid_dtype(self.dtype):
            raise ValueError("dtype must be a valid dtype for numpy")

    def encode_example(self, example):
        if not isinstance(example, np.ndarray):
            example = np.asarray(example, dtype=self.dtype)
        # Ensure the shape and dtype match
        if example.dtype != self.dtype:
            raise ValueError(
                "Dtype {} do not match {}".format(example.dtype, self.dtype)
            )
        assert_shape_match(example.shape, self.shape)
        return example.tobytes()

    def decode_example(self, example):
        if isinstance(example, StreamWrapper):
            fobj = example
            example = example.read()
            fobj.close()
            if not example:
                return None
        shape = [-1 if dim is None else dim for dim in self.shape]
        return np.frombuffer(example, dtype=self.dtype).reshape(shape)


class Dimension:
    __slots__ = ["_value"]

    def __init__(self, value):
        """Creates a new Dimension with the given value."""
        if isinstance(value, int):  # Most common case.
            if value < 0:
                raise ValueError("Dimension %d must be >= 0" % value)
            self._value = value
        elif value is None:
            self._value = None
        elif isinstance(value, Dimension):
            self._value = value._value
        else:
            try:
                self._value = value.__index__()
            except AttributeError:
                raise TypeError(
                    "Dimension value must be integer or None or have "
                    "an __index__ method, got value '{0!r}' with type '{1!r}'".format(
                        value, type(value)
                    )
                )
            if self._value < 0:
                raise ValueError("Dimension %d must be >= 0" % self._value)

    def __repr__(self):
        return "Dimension(%s)" % repr(self._value)

    def __str__(self):
        value = self._value
        return "?" if value is None else str(value)

    def __eq__(self, other):
        try:
            other = as_dimension(other)
        except (TypeError, ValueError):
            return NotImplemented
        if self._value is None or other.value is None:
            return None
        return self._value == other.value

    def __ne__(self, other):
        try:
            other = as_dimension(other)
        except (TypeError, ValueError):
            return NotImplemented
        if self._value is None or other.value is None:
            return None
        return self._value != other.value

    def __int__(self):
        return self._value

    def __index__(self):
        return self._value

    @property
    def value(self):
        return self._value

    def is_compatible_with(self, other):
        other = as_dimension(other)
        return self._value is None or other.value is None or self._value == other.value

    def assert_is_compatible_with(self, other):
        if not self.is_compatible_with(other):
            raise ValueError("Dimensions %s and %s are not compatible" % (self, other))

    def merge_with(self, other):
        other = as_dimension(other)
        self.assert_is_compatible_with(other)
        if self._value is None:
            return Dimension(other.value)
        else:
            return Dimension(self._value)

    def __add__(self, other):
        try:
            other = as_dimension(other)
        except (TypeError, ValueError):
            return NotImplemented
        if self._value is None or other.value is None:
            return Dimension(None)
        else:
            return Dimension(self._value + other.value)

    def __radd__(self, other):
        return self + other

    def __sub__(self, other):
        try:
            other = as_dimension(other)
        except (TypeError, ValueError):
            return NotImplemented
        if self._value is None or other.value is None:
            return Dimension(None)
        else:
            return Dimension(self._value - other.value)

    def __rsub__(self, other):
        other = as_dimension(other)
        if self._value is None or other.value is None:
            return Dimension(None)
        else:
            return Dimension(other.value - self._value)

    def __mul__(self, other):
        try:
            other = as_dimension(other)
        except (TypeError, ValueError):
            return NotImplemented

        if self._value is None or other.value is None:
            return Dimension(None)
        else:
            return Dimension(self._value * other.value)

    def __rmul__(self, other):
        return self * other

    def __floordiv__(self, other):
        try:
            other = as_dimension(other)
        except (TypeError, ValueError):
            return NotImplemented
        if self._value is None or other.value is None:
            return Dimension(None)
        else:
            return Dimension(self._value // other.value)

    def __rfloordiv__(self, other):
        other = as_dimension(other)
        if self._value is None or other.value is None:
            return Dimension(None)
        else:
            return Dimension(other.value // self._value)

    def __truediv__(self, other):
        raise TypeError(
            "unsupported operand type(s) for /: 'Dimension' and '{}', "
            "please use // instead".format(type(other).__name__)
        )

    def __rtruediv__(self, other):
        raise TypeError(
            "unsupported operand type(s) for /: '{}' and 'Dimension', "
            "please use // instead".format(type(other).__name__)
        )

    def __mod__(self, other):
        other = as_dimension(other)
        if self._value is None or other.value is None:
            return Dimension(None)
        else:
            return Dimension(self._value % other.value)

    def __rmod__(self, other):
        other = as_dimension(other)
        return other % self

    def __lt__(self, other):
        other = as_dimension(other)
        if self._value is None or other.value is None:
            return None
        else:
            return self._value < other.value

    def __le__(self, other):
        other = as_dimension(other)
        if self._value is None or other.value is None:
            return None
        else:
            return self._value <= other.value

    def __gt__(self, other):
        other = as_dimension(other)
        if self._value is None or other.value is None:
            return None
        else:
            return self._value > other.value

    def __ge__(self, other):
        other = as_dimension(other)
        if self._value is None or other.value is None:
            return None
        else:
            return self._value >= other.value

    def __reduce__(self):
        return Dimension, (self._value,)


def as_dimension(value):
    if isinstance(value, Dimension):
        return value
    else:
        return Dimension(value)


class TensorShape:
    def __init__(self, dims):
        if isinstance(dims, (tuple, list)):  # Most common case.
            self._dims = [Dimension(d) for d in dims]
        elif dims is None:
            self._dims = None
        elif isinstance(dims, TensorShape):
            self._dims = dims.dims
        else:
            try:
                dims_iter = iter(dims)
            except TypeError:
                # Treat as a singleton dimension
                self._dims = [as_dimension(dims)]
            else:
                self._dims = []
                for d in dims_iter:
                    try:
                        self._dims.append(as_dimension(d))
                    except TypeError as e:
                        raise TypeError(
                            "Failed to convert '{0!r}' to a shape: '{1!r}'"
                            "could not be converted to a dimension. A shape should "
                            "either be single dimension (e.g. 10), or an iterable of "
                            "dimensions (e.g. [1, 10, None]).".format(dims, d)
                        ) from e

    def __repr__(self):
        return "TensorShape(%r)" % self._dims

    def __str__(self):
        if self.rank is None:
            return "<unknown>"
        elif self.rank == 1:
            return "(%s,)" % self._dims[0]
        else:
            return "(%s)" % ", ".join(str(d) for d in self._dims)

    @property
    def rank(self):
        if self._dims is not None:
            return len(self._dims)
        return None

    @property
    def dims(self):
        return self._dims

    @property
    def ndims(self):
        return self.rank

    def __len__(self):
        if self._dims is None:
            raise ValueError("Cannot take the length of shape with unknown rank.")
        return len(self._dims)

    def __bool__(self):
        return self._dims is not None

    def __iter__(self):
        if self._dims is None:
            raise ValueError("Cannot iterate over a shape with unknown rank.")
        else:
            return iter(d for d in self._dims)

    def __getitem__(self, key):
        if self._dims is not None:
            if isinstance(key, slice):
                return TensorShape(self._dims[key])
            else:
                return self._dims[key]
        else:
            if isinstance(key, slice):
                start = key.start if key.start is not None else 0
                stop = key.stop

                if key.step is not None:
                    raise ValueError("Steps are not yet handled")
                if stop is None:
                    return unknown_shape()
                elif start < 0 or stop < 0:
                    return unknown_shape()
                else:
                    return unknown_shape(rank=stop - start)
            else:
                return Dimension(None)

    def num_elements(self):
        if self.is_fully_defined():
            size = 1
            for dim in self._dims:
                size *= dim.value
            return size
        else:
            return None

    def merge_with(self, other):
        other = as_shape(other)
        if self._dims is None:
            return other
        else:
            try:
                self.assert_same_rank(other)
                new_dims = []
                for i, dim in enumerate(self._dims):
                    new_dims.append(dim.merge_with(other[i]))
                return TensorShape(new_dims)
            except ValueError:
                raise ValueError("Shapes %s and %s are not compatible" % (self, other))

    def __add__(self, other):
        if not isinstance(other, TensorShape):
            other = TensorShape(other)
        return self.concatenate(other)

    def __radd__(self, other):
        if not isinstance(other, TensorShape):
            other = TensorShape(other)
        return other.concatenate(self)

    def concatenate(self, other):
        other = as_shape(other)
        if self._dims is None or other.dims is None:
            return unknown_shape()
        else:
            return TensorShape(self._dims + other.dims)

    def assert_same_rank(self, other):
        other = as_shape(other)
        if self.rank is not None and other.rank is not None:
            if self.rank != other.rank:
                raise ValueError(
                    "Shapes %s and %s must have the same rank" % (self, other)
                )

    def assert_has_rank(self, rank):
        if self.rank not in (None, rank):
            raise ValueError("Shape %s must have rank %d" % (self, rank))

    def with_rank(self, rank):
        try:
            return self.merge_with(unknown_shape(rank=rank))
        except ValueError:
            raise ValueError("Shape %s must have rank %d" % (self, rank))

    def with_rank_at_least(self, rank):
        if self.rank is not None and self.rank < rank:
            raise ValueError("Shape %s must have rank at least %d" % (self, rank))
        else:
            return self

    def with_rank_at_most(self, rank):
        if self.rank is not None and self.rank > rank:
            raise ValueError("Shape %s must have rank at most %d" % (self, rank))
        else:
            return self

    def is_compatible_with(self, other):
        other = as_shape(other)
        if self._dims is not None and other.dims is not None:
            if self.rank != other.rank:
                return False
            for x_dim, y_dim in zip(self._dims, other.dims):
                if not x_dim.is_compatible_with(y_dim):
                    return False
        return True

    def assert_is_compatible_with(self, other):
        if not self.is_compatible_with(other):
            raise ValueError("Shapes %s and %s are incompatible" % (self, other))

    def most_specific_compatible_shape(self, other):
        other = as_shape(other)
        if self._dims is None or other.dims is None or self.rank != other.rank:
            return unknown_shape()

        dims = [(Dimension(None))] * self.rank
        for i, (d1, d2) in enumerate(zip(self._dims, other.dims)):
            if d1 is not None and d2 is not None and d1 == d2:
                dims[i] = d1
        return TensorShape(dims)

    def is_fully_defined(self):
        return self._dims is not None and all(
            dim.value is not None for dim in self._dims
        )

    def assert_is_fully_defined(self):
        if not self.is_fully_defined():
            raise ValueError("Shape %s is not fully defined" % self)

    def as_list(self):
        if self._dims is None:
            raise ValueError("as_list() is not defined on an unknown TensorShape.")
        return [dim.value for dim in self._dims]

    def __eq__(self, other):
        """Returns True if `self` is equivalent to `other`."""
        try:
            other = as_shape(other)
        except TypeError:
            return NotImplemented
        return self._dims == other.dims

    def __ne__(self, other):
        try:
            other = as_shape(other)
        except TypeError:
            return NotImplemented
        if self.rank is None or other.rank is None:
            raise ValueError("The inequality of unknown TensorShapes is undefined.")
        if self.rank != other.rank:
            return True
        return self._dims != other.dims

    def __reduce__(self):
        return TensorShape, (self._dims,)

    def __concat__(self, other):
        return self.concatenate(other)


def as_shape(shape):
    if isinstance(shape, TensorShape):
        return shape
    else:
        return TensorShape(shape)


def unknown_shape(rank=None, **kwargs):
    if rank is None and "ndims" in kwargs:
        rank = kwargs.pop("ndims")
    if kwargs:
        raise TypeError("Unknown argument: %s" % kwargs)
    if rank is None:
        return TensorShape(None)
    else:
        return TensorShape([Dimension(None)] * rank)


def assert_shape_match(shape1, shape2):
    """Ensure the shape1 match the pattern given by shape2.

    Args:
      shape1 (tuple): Static shape
      shape2 (tuple): Dynamic shape (can contain None)
    """
    shape1 = TensorShape(shape1)
    shape2 = TensorShape(shape2)
    if shape1.ndims is None or shape2.ndims is None:
        raise ValueError(
            "Shapes must have known rank. Got %s and %s." % (shape1.ndims, shape2.ndims)
        )
    shape1.assert_same_rank(shape2)
    shape1.assert_is_compatible_with(shape2)


def is_valid_dtype(dtype_str):
    try:
        np.dtype(dtype_str)
        return True
    except TypeError:
        return False



import copy
from dataclasses import fields
from functools import wraps
from typing import Any, Dict

from tarzan import features
from tarzan.features.sequence import Sequence
from tarzan.utils import asdict, zip_dict


def encode_nested_example(schema, obj, level=0):
    """Encode a nested example.
    This is used since some features (in particular ClassLabel) have some logic during encoding.

    To avoid iterating over possibly long lists, it first checks (recursively) if the first element that is not None or
    empty (if it is a sequence) has to be encoded. If the first element needs to be encoded, then all the elements of
    the list will be encoded, otherwise they'll stay the same.
    """
    # Nested structures: we allow dict, list/tuples, sequences
    if isinstance(schema, dict):
        if level == 0 and obj is None:
            raise ValueError("Got None but expected a dictionary instead")
        return (
            {
                k: encode_nested_example(sub_schema, sub_obj, level=level + 1)
                for k, (sub_schema, sub_obj) in zip_dict(schema, obj)
            }
            if obj is not None
            else None
        )

    elif isinstance(schema, (list, tuple)):
        sub_schema = schema[0]
        if obj is None:
            return None
        else:
            if len(obj) > 0:
                return [
                    encode_nested_example(sub_schema, o, level=level + 1) for o in obj
                ]
            return list(obj)
    elif isinstance(schema, Sequence):
        if obj is None:
            return None
        # We allow to reverse list of dict => dict of list for compatibility with tfds
        if isinstance(schema.feature, dict):
            # dict of list to fill
            list_dict = {}
            if isinstance(obj, (list, tuple)):
                # obj is a list of dict
                for k, dict_tuples in zip_dict(schema.feature, *obj):
                    list_dict[k] = [
                        encode_nested_example(dict_tuples[0], o, level=level + 1)
                        for o in dict_tuples[1:]
                    ]
                return list_dict
            else:
                # obj is a single dict
                for k, (sub_schema, sub_objs) in zip_dict(schema.feature, obj):
                    list_dict[k] = [
                        encode_nested_example(sub_schema, o, level=level + 1)
                        for o in sub_objs
                    ]
                return list_dict
        # schema.feature is not a dict
        if isinstance(obj, str):  # don't interpret a string as a list
            raise ValueError(f"Got a string but expected a list instead: '{obj}'")
        else:
            if len(obj) > 0:
                return [
                    encode_nested_example(schema.feature, o, level=level + 1)
                    for o in obj
                ]
            return list(obj)
    # Object with special encoding:
    else:
        return schema.encode_example(obj) if obj is not None else None


def decode_nested_example(schema, obj):
    """Decode a nested example.
    This is used since some features (in particular Audio and Image) have some logic during decoding.

    """
    # Nested structures: we allow dict, list/tuples, sequences
    if isinstance(schema, dict):
        return (
            {
                k: decode_nested_example(sub_schema, sub_obj)
                for k, (sub_schema, sub_obj) in zip_dict(schema, obj)
            }
            if obj is not None
            else None
        )
    elif isinstance(schema, (list, tuple)):
        sub_schema = schema[0]
        if obj is None:
            return None
        else:
            if len(obj) > 0:
                return [decode_nested_example(sub_schema, o) for o in obj]
            return list(obj)
    elif isinstance(schema, Sequence):
        # We allow to reverse list of dict => dict of list for compatibility with tfds
        if isinstance(schema.feature, dict):
            return {
                k: decode_nested_example([schema.feature[k]], obj[k])
                for k in schema.feature
            }
        else:
            return decode_nested_example([schema.feature], obj)
    # Object with special decoding:
    else:
        # we pass the token to read and decode files from private repositories in streaming mode
        return schema.decode_example(obj) if obj is not None else None


def generate_from_dict(obj: Any):
    """Regenerate the nested feature object from a deserialized dict.
    We use the '_type' fields to get the dataclass name to load.

    generate_from_dict is the recursive helper for Features.from_dict, and allows for a convenient constructor syntax
    to define features from deserialized JSON dictionaries. This function is used in particular when deserializing
    a :class:`DatasetInfo` that was dumped to a JSON object.
    """
    # Nested structures: we allow dict, list/tuples, sequences
    if isinstance(obj, list):
        return [generate_from_dict(value) for value in obj]
    # Otherwise we have a dict or a dataclass
    if "_type" not in obj or isinstance(obj["_type"], dict):
        return {key: generate_from_dict(value) for key, value in obj.items()}
    class_type = getattr(features, obj.pop("_type"))

    if class_type == Sequence:
        return Sequence(
            feature=generate_from_dict(obj["feature"]), length=obj.get("length", -1)
        )

    field_names = {f.name for f in fields(class_type)}
    return class_type(**{k: v for k, v in obj.items() if k in field_names})


def keep_features_dicts_synced(func):
    """
    Wrapper to keep the secondary dictionary, which tracks whether keys are decodable, of the :class:`datasets.Features`
    object in sync with the main dictionary.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        if args:
            self: "Features" = args[0]
            args = args[1:]
        else:
            self: "Features" = kwargs.pop("self")
        out = func(self, *args, **kwargs)
        return out

    wrapper._decorator_name_ = "_keep_dicts_synced"
    return wrapper


class Features(dict):
    """A special dictionary that defines the internal structure of a dataset."""

    def __init__(*args, **kwargs):
        # self not in the signature to allow passing self as a kwarg
        if not args:
            raise TypeError(
                "descriptor '__init__' of 'Features' object needs an argument"
            )
        self, *args = args
        super(Features, self).__init__(*args, **kwargs)

    __setitem__ = keep_features_dicts_synced(dict.__setitem__)
    __delitem__ = keep_features_dicts_synced(dict.__delitem__)
    update = keep_features_dicts_synced(dict.update)
    setdefault = keep_features_dicts_synced(dict.setdefault)
    pop = keep_features_dicts_synced(dict.pop)
    popitem = keep_features_dicts_synced(dict.popitem)
    clear = keep_features_dicts_synced(dict.clear)

    def __reduce__(self):
        return Features, (dict(self),)

    @classmethod
    def from_dict(cls, dic) -> "Features":
        """
        Construct [`Features`] from dict.

        Regenerate the nested feature object from a deserialized dict.
        We use the `_type` key to infer the dataclass name of the feature `FieldType`.

        It allows for a convenient constructor syntax
        to define features from deserialized JSON dictionaries. This function is used in particular when deserializing
        a [`DatasetInfo`] that was dumped to a JSON object.

        Args:
            dic (`dict[str, Any]`):
                Python dictionary.

        Returns:
            `Features`

        Example::
            >>> Features.from_dict({'_type': {'dtype': 'string', 'id': None, '_type': 'Value'}})
            {'_type': Value(dtype='string', id=None)}
        """
        obj = generate_from_dict(dic)
        return cls(**obj)

    def to_dict(self):
        return asdict(self)

    def encode_example(self, example):
        """
        Encode example into a format for storage.

        Args:
            example (`dict[str, Any]`):
                Data in a Dataset row.

        Returns:
            `dict[str, Any]`
        """
        return encode_nested_example(self, example)

    def encode_column(self, column, column_name: str):
        """
        Encode column into a format for storage.

        Args:
            column (`list[Any]`):
                Data in a Dataset column.
            column_name (`str`):
                Dataset column name.

        Returns:
            `list[Any]`
        """
        return [encode_nested_example(self[column_name], obj) for obj in column]

    def encode_batch(self, batch):
        """
        Encode batch into a format for storage.

        Args:
            batch (`dict[str, list[Any]]`):
                Data in a Dataset batch.

        Returns:
            `dict[str, list[Any]]`
        """
        encoded_batch = {}
        if set(batch) != set(self):
            raise ValueError(
                f"Column mismatch between batch {set(batch)} and features {set(self)}"
            )
        for key, column in batch.items():
            encoded_batch[key] = [
                encode_nested_example(self[key], obj) for obj in column
            ]
        return encoded_batch

    def decode_example(self, example: Dict):
        """Decode example with custom feature decoding.

        Args:
            example (`dict[str, Any]`):
                Dataset row data.

        Returns:
            `dict[str, Any]`
        """

        return {
            column_name: decode_nested_example(feature, value)
            for column_name, (feature, value) in zip_dict(
                {key: value for key, value in self.items() if key in example}, example
            )
        }

    def decode_column(self, column: list, column_name: str):
        """Decode column with custom feature decoding.

        Args:
            column (`list[Any]`):
                Dataset column data.
            column_name (`str`):
                Dataset column name.

        Returns:
            `list[Any]`
        """
        return [
            (
                decode_nested_example(self[column_name], value)
                if value is not None
                else None
            )
            for value in column
        ]

    def decode_batch(self, batch: Dict):
        """Decode batch with custom feature decoding.

        Args:
            batch (`dict[str, list[Any]]`):
                Dataset batch data.

        Returns:
            `dict[str, list[Any]]`
        """
        decoded_batch = {}
        for column_name, column in batch.items():
            decoded_batch[column_name] = [
                (
                    decode_nested_example(self[column_name], value)
                    if value is not None
                    else None
                )
                for value in column
            ]
        return decoded_batch

    def copy(self) -> "Features":
        """
        Make a deep copy of [`Features`].

        Returns:
            [`Features`]
        """
        return copy.deepcopy(self)
